from typing import Union
import warnings
from abc import abstractmethod

import pandas as pd

from optimus.engines.base.meta import Meta
from optimus.helpers.types import *
from optimus.infer import is_dict, is_tuple


class BaseCreate:
    def __init__(self, op: 'DataFrameType'):
        self.op = op

    def _dictionary(self, dict, force_dtypes=False):

        new_dict = {}

        for key, values in dict.items():
            if is_tuple(key):
                dtype = None
                force_dtype = force_dtypes
                nulls = False
                if len(key) == 4:
                    name, dtype, nulls, force_dtype = key
                if len(key) == 3:
                    name, dtype, nulls = key
                elif len(key) == 2:
                    name, dtype = key
                if force_dtype:
                    dtype = self.op.constants.OPTIMUS_TO_INTERNAL.get(dtype, dtype)
            else:
                name = key
                dtype = None
                nulls = False
                force_dtype = force_dtypes

            new_dict[(name, dtype, nulls, force_dtype)] = values

        return new_dict

    @property
    def _pd(self):
        return pd

    def _dfd_from_dict(self, dict) -> 'InternalDataFrameType':
        pd_dict = {}
        for (name, dtype, nulls, force_dtype), values in dict.items():
            dtype = self.op.constants.COMPATIBLE_DTYPES.get(dtype, dtype) if force_dtype else None
            pd_series = self._pd.Series(values, dtype=dtype)
            pd_dict.update({name: pd_series})
        return self._pd.DataFrame(pd_dict)

    @abstractmethod
    def _df_from_dfd(self, dfd, *args, **kwargs) -> 'DataFrameType':
        pass

    def dataframe(self, data: Union[dict, 'InternalDataFrameType'] = None, force_data_types=False,
                  n_partitions: int = 1, *args, **kwargs) -> 'DataFrameType':
        """Creates a dataframe using a dictionary or a Pandas DataFrame

        Creates a dataframe using the form
        `{"Column name": ["value 1", "value 2"], ...}` or 
        `{("Column name", "str", True, True): ["value 1", "value 2"]}`,
        where the tuple uses the form `(str, str, boolean, boolean)` for 
        `(name, data type, allow nulls, force data type in creation)`. You can
        also pass 2-length and 3-length tuples.
        :param data: A pandas dataframe or dictionary to construct the dataframe.
        :param force_data_types: Force every data type passed to data.
        :param n_partitions: Number of partitions (For distributed engines only)
        :return: BaseDataFrame
        """

        if data is None and len(kwargs):
            data = kwargs
            kwargs = {}

        data_dict = None

        if is_dict(data):
            data_dict = self._dictionary(data, force_dtypes=force_data_types)
            data = self._dfd_from_dict(data_dict)

        df = self._df_from_dfd(data, n_partitions=n_partitions, *args, **kwargs)

        try:
            df.meta = Meta.set(df.meta, value={"max_cell_length": df.cols.len("*").cols.max()})
        except:
            warnings.warn("Could not set max_cell_length")

        if data_dict is not None:
            for (name, dtype, nulls, force_dtype) in data_dict:
                if dtype and not force_dtype:
                    df = df.cols.set_data_type(name, dtype)

        return df



import datetime
import numpy as np
from optimus.tests.base import TestBase
from optimus.helpers.json import json_encoding
from optimus.helpers.functions import deep_sort, df_dicts_equal, results_equal


def Timestamp(t):
    return datetime.datetime.strptime(t, "%Y-%m-%d %H:%M:%S")


NaT = np.datetime64('NaT')
nan = float("nan")
inf = float("inf")


class TestMathPandas(TestBase):
    config = {'engine': 'pandas'}
    dict = {('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan]}
    maxDiff = None

    def test_cols_add(self):
        df = self.create_dataframe(data={('add_test1', 'float64'): [nan, nan, -9.0, 10.234, -inf, -42.0], ('add_test2', 'float64'): [nan, 1.0, 9.0, 703.0, inf, -321.0]}, force_data_types=True)
        result = df.cols.add(cols=['add_test1', 'add_test2'])
        expected = self.create_dataframe(data={('add_test1', 'float64'): [nan, nan, -9.0, 10.234, -inf, -42.0], ('add_test2', 'float64'): [nan, 1.0, 9.0, 703.0, inf, -321.0], ('add_add_test1_add_test2', 'float64'): [nan, nan, 0.0, 713.234, nan, -363.0]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_add_2_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank'])
        result = df.cols.add(cols=['height(ft)', 'rank'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('add_height(ft)_rank', 'float64'): [-18.0, 24.0, 33.0, 21.0, nan, 308.0]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_add_3_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank', 'age'])
        result = df.cols.add(cols=['height(ft)', 'rank', 'age'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('add_height(ft)_rank_age', 'float64'): [4999982.0, 5000024.0, 5000033.0, 5000021.0, nan, 5000308.0]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_add_all(self):
        df = self.df.copy()
        result = df.cols.add(cols='*')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('add_*', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_add_all_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank', 'age', 'weight(t)'])
        result = df.cols.add(cols=['height(ft)', 'rank', 'age', 'weight(t)'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('add_height(ft)_rank_age_weight(t)', 'float64'): [4999986.3, 5000026.0, 5000037.0, 5000022.8, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_add_string(self):
        df = self.df.copy().cols.select(['names', 'function'])
        result = df.cols.add(cols=['names', 'function'], output_col='names-function')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names-function', 'object'): ['OptimusLeader', 'bumbl#ebéé  Espionage', 'ironhide&Security', 'JazzFirst Lieutenant', 'MegatronNone', 'Metroplex_)^$Battle Station']}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_add_various_types(self):
        df = self.df.copy()
        result = df.cols.add(cols=['NullType', 'weight(t)', 'japanese name'], output_col='nt+wt+jn')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('nt+wt+jn', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_div(self):
        df = self.create_dataframe(data={('div_test1', 'float64'): [nan, nan, -8.0, 10.234, -inf, -42.0], ('div_test2', 'float64'): [nan, 1.0, 0.0, 703.0, inf, -321.0]}, force_data_types=True)
        result = df.cols.div(cols=['div_test1', 'div_test2'])
        expected = self.create_dataframe(data={('div_test1', 'float64'): [nan, nan, -8.0, 10.234, -inf, -42.0], ('div_test2', 'float64'): [nan, 1.0, 0.0, 703.0, inf, -321.0], ('div_div_test1_div_test2', 'float64'): [nan, nan, -inf, 0.014557610241820769, nan, 0.1308411214953271]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_div_2_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank'])
        result = df.cols.div(cols=['height(ft)', 'rank'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('div_height(ft)_rank', 'float64'): [-2.8, 2.4285714285714284, 3.7142857142857144, 1.625, nan, 37.5]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_div_3_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank', 'age'])
        result = df.cols.div(cols=['height(ft)', 'rank', 'age'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('div_height(ft)_rank_age', 'float64'): [-5.599999999999999e-07, 4.857142857142856e-07, 7.428571428571429e-07, 3.25e-07, nan, 7.5e-06]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_div_all(self):
        df = self.df.copy()
        result = df.cols.div(cols='*')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('div_*', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_div_all_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank', 'age', 'weight(t)'])
        result = df.cols.div(cols=['height(ft)', 'rank', 'age', 'weight(t)'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('div_height(ft)_rank_age_weight(t)', 'float64'): [-1.3023255813953487e-07, 2.428571428571428e-07, 1.8571428571428572e-07, 1.8055555555555556e-07, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_div_string(self):
        df = self.df.copy().cols.select(['names', 'function'])
        result = df.cols.div(cols=['names', 'function'], output_col='names/function')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names/function', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_div_various_types(self):
        df = self.df.copy()
        result = df.cols.div(cols=['NullType', 'weight(t)', 'japanese name'], output_col='nt*wt*jn')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('nt*wt*jn', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_mul(self):
        df = self.create_dataframe(data={('mul_test1', 'float64'): [nan, nan, 8.0, 10.234, -inf, -42.0], ('mul_test2', 'float64'): [nan, 1.0, 0.125, 703.0, inf, -321.0]}, force_data_types=True)
        result = df.cols.mul(cols=['mul_test1', 'mul_test2'])
        expected = self.create_dataframe(data={('mul_test1', 'float64'): [nan, nan, 8.0, 10.234, -inf, -42.0], ('mul_test2', 'float64'): [nan, 1.0, 0.125, 703.0, inf, -321.0], ('mul_mul_test1_mul_test2', 'float64'): [nan, nan, 1.0, 7194.502, -inf, 13482.0]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_mul_2_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank'])
        result = df.cols.mul(cols=['height(ft)', 'rank'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('mul_height(ft)_rank', 'float64'): [-280.0, 119.0, 182.0, 104.0, nan, 2400.0]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_mul_3_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank', 'age'])
        result = df.cols.mul(cols=['height(ft)', 'rank', 'age'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('mul_height(ft)_rank_age', 'float64'): [-1400000000.0, 595000000.0, 910000000.0, 520000000.0, nan, 12000000000.0]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_mul_all(self):
        df = self.df.copy()
        result = df.cols.mul(cols='*')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('mul_*', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_mul_all_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank', 'age', 'weight(t)'])
        result = df.cols.mul(cols=['height(ft)', 'rank', 'age', 'weight(t)'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('mul_height(ft)_rank_age_weight(t)', 'float64'): [-6020000000.0, 1190000000.0, 3640000000.0, 936000000.0, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_mul_string(self):
        df = self.df.copy().cols.select(['names', 'function'])
        result = df.cols.mul(cols=['names', 'function'], output_col='names*function')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names*function', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_mul_various_types(self):
        df = self.df.copy()
        result = df.cols.mul(cols=['NullType', 'weight(t)', 'japanese name'], output_col='nt*wt*jn')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('nt*wt*jn', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_rdiv(self):
        df = self.create_dataframe(data={('rdiv_test1', 'float64'): [nan, nan, -8.0, 10.234, -inf, -42.0], ('rdiv_test2', 'float64'): [nan, 1.0, 0.0, 703.0, inf, -321.0]}, force_data_types=True)
        result = df.cols.rdiv(cols=['rdiv_test1', 'rdiv_test2'])
        expected = self.create_dataframe(data={('rdiv_test1', 'float64'): [nan, nan, -8.0, 10.234, -inf, -42.0], ('rdiv_test2', 'float64'): [nan, 1.0, 0.0, 703.0, inf, -321.0], ('rdiv_rdiv_test1_rdiv_test2', 'float64'): [nan, nan, -0.0, 68.69259331639633, nan, 7.642857142857143]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_rdiv_2_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank'])
        result = df.cols.rdiv(cols=['height(ft)', 'rank'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('rdiv_height(ft)_rank', 'float64'): [-0.35714285714285715, 0.4117647058823529, 0.2692307692307692, 0.6153846153846154, nan, 0.02666666666666667]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_rdiv_3_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank', 'age'])
        result = df.cols.rdiv(cols=['height(ft)', 'rank', 'age'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('rdiv_height(ft)_rank_age', 'float64'): [-14000000.0, 12142857.142857144, 18571428.57142857, 8125000.0, nan, 187500000.0]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_rdiv_all(self):
        df = self.df.copy()
        result = df.cols.rdiv(cols='*')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('rdiv_*', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_rdiv_all_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank', 'age', 'weight(t)'])
        result = df.cols.rdiv(cols=['height(ft)', 'rank', 'age', 'weight(t)'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('rdiv_height(ft)_rank_age_weight(t)', 'float64'): [-3.071428571428571e-07, 1.6470588235294117e-07, 2.153846153846154e-07, 2.2153846153846153e-07, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_rdiv_string(self):
        df = self.df.copy().cols.select(['names', 'function'])
        result = df.cols.rdiv(cols=['names', 'function'], output_col='names*function')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names*function', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_rdiv_various_types(self):
        df = self.df.copy()
        result = df.cols.rdiv(cols=['NullType', 'weight(t)', 'japanese name'], output_col='nt*wt*jn')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('nt*wt*jn', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_sub(self):
        df = self.create_dataframe(data={('sub_test1', 'float64'): [nan, nan, 9.0, 10.234, inf, -42.0], ('sub_test2', 'float64'): [nan, 1.0, 9.0, 703.0, inf, -321.0]}, force_data_types=True)
        result = df.cols.sub(cols=['sub_test1', 'sub_test2'])
        expected = self.create_dataframe(data={('sub_test1', 'float64'): [nan, nan, 9.0, 10.234, inf, -42.0], ('sub_test2', 'float64'): [nan, 1.0, 9.0, 703.0, inf, -321.0], ('sub_sub_test1_sub_test2', 'float64'): [nan, nan, 0.0, -692.766, nan, 279.0]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_sub_2_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank'])
        result = df.cols.sub(cols=['height(ft)', 'rank'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('sub_height(ft)_rank', 'float64'): [-38.0, 10.0, 19.0, 5.0, nan, 292.0]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_sub_3_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank', 'age'])
        result = df.cols.sub(cols=['height(ft)', 'rank', 'age'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('sub_height(ft)_rank_age', 'float64'): [-5000038.0, -4999990.0, -4999981.0, -4999995.0, nan, -4999708.0]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_sub_all(self):
        df = self.df.copy()
        result = df.cols.sub(cols='*')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('sub_*', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_sub_all_numerics(self):
        df = self.df.copy().cols.select(['height(ft)', 'rank', 'age', 'weight(t)'])
        result = df.cols.sub(cols=['height(ft)', 'rank', 'age', 'weight(t)'])
        expected = self.create_dataframe(data={('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('sub_height(ft)_rank_age_weight(t)', 'float64'): [-5000042.3, -4999992.0, -4999985.0, -4999996.8, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_sub_string(self):
        df = self.df.copy().cols.select(['names', 'function'])
        result = df.cols.sub(cols=['names', 'function'], output_col='names-function')
        expected = self.create_dataframe(data={('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names-function', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_sub_various_types(self):
        df = self.df.copy()
        result = df.cols.sub(cols=['NullType', 'weight(t)', 'japanese name'], output_col='nt-wt-jn')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan], ('nt-wt-jn', 'float64'): [nan, nan, nan, nan, nan, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))


class TestMathDask(TestMathPandas):
    config = {'engine': 'dask', 'n_partitions': 1}


class TestMathPartitionDask(TestMathPandas):
    config = {'engine': 'dask', 'n_partitions': 2}


try:
    import cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestMathCUDF(TestMathPandas):
        config = {'engine': 'cudf'}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestMathDC(TestMathPandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 1}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestMathPartitionDC(TestMathPandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 2}


try:
    import pyspark # pyright: reportMissingImports=false
except:
    pass
else:
    class TestMathSpark(TestMathPandas):
        config = {'engine': 'spark'}


try:
    import vaex # pyright: reportMissingImports=false
except:
    pass
else:
    class TestMathVaex(TestMathPandas):
        config = {'engine': 'vaex'}



from optimus.helpers.constants import ProfilerDataTypes
import numpy as np


class BaseConstants:

    # inferred/input to internal
    OPTIMUS_TO_INTERNAL = {ProfilerDataTypes.INT.value: "int",
                           ProfilerDataTypes.FLOAT.value: "float",
                           ProfilerDataTypes.STRING.value: "str",
                           ProfilerDataTypes.BOOL.value: "bool",
                           ProfilerDataTypes.BOOLEAN.value: "bool",
                           ProfilerDataTypes.DATETIME.value: "datetime64[ns]",
                           ProfilerDataTypes.ARRAY.value: "object",
                           ProfilerDataTypes.OBJECT.value: "object",
                           ProfilerDataTypes.GENDER.value: "gender",
                           ProfilerDataTypes.IP.value: "str",
                           ProfilerDataTypes.URL.value: "str",
                           ProfilerDataTypes.EMAIL.value: "str",
                           ProfilerDataTypes.CREDIT_CARD_NUMBER.value: "str",
                           ProfilerDataTypes.ZIP_CODE.value: "object",
                           ProfilerDataTypes.MISSING.value: "object",
                           ProfilerDataTypes.PHONE_NUMBER.value: "str",
                           ProfilerDataTypes.SOCIAL_SECURITY_NUMBER.value: "str",
                           ProfilerDataTypes.HTTP_CODE.value: "str",
                           ProfilerDataTypes.US_STATE.value: "category",
                           ProfilerDataTypes.CATEGORICAL.value: "category",
                           ProfilerDataTypes.NULL.value: "object",
                           "double": "float",
                           "vector": "object",
                           "list": "object",
                           "dict": "object",
                           "date": "datetime64[ns]",
                           "time": "datetime64[ns]",
                           "timestamp": "datetime64[ns]"}

    # internal to inferred
    INTERNAL_TO_OPTIMUS = {"string_": ProfilerDataTypes.STRING.value,
                           "large_string": ProfilerDataTypes.STRING.value,
                           "unicode_": ProfilerDataTypes.STRING.value,
                           "string": ProfilerDataTypes.STRING.value,
                           "int_": ProfilerDataTypes.INT.value,
                           "bigint": ProfilerDataTypes.INT.value,
                           "int8": ProfilerDataTypes.INT.value,
                           "int16": ProfilerDataTypes.INT.value,
                           "int32": ProfilerDataTypes.INT.value,
                           "int64": ProfilerDataTypes.INT.value,
                           "uint": ProfilerDataTypes.INT.value,
                           "uint8": ProfilerDataTypes.INT.value,
                           "uint16": ProfilerDataTypes.INT.value,
                           "uint32": ProfilerDataTypes.INT.value,
                           "uint64": ProfilerDataTypes.INT.value,
                           "binary": ProfilerDataTypes.INT.value,
                           "large_binary": ProfilerDataTypes.INT.value,
                           "numeric": ProfilerDataTypes.FLOAT.value,
                           "float": ProfilerDataTypes.FLOAT.value,
                           "float16": ProfilerDataTypes.FLOAT.value,
                           "float32": ProfilerDataTypes.FLOAT.value,
                           "float64": ProfilerDataTypes.FLOAT.value,
                           "float_": ProfilerDataTypes.FLOAT.value,
                           "double": ProfilerDataTypes.FLOAT.value,
                           "bool_": ProfilerDataTypes.BOOL.value,
                           "date": ProfilerDataTypes.DATETIME.value,
                           "date32": ProfilerDataTypes.DATETIME.value,
                           "date64": ProfilerDataTypes.DATETIME.value,
                           "time": ProfilerDataTypes.DATETIME.value,
                           "time32": ProfilerDataTypes.DATETIME.value,
                           "time64": ProfilerDataTypes.DATETIME.value,
                           "timestamp": ProfilerDataTypes.DATETIME.value,
                           "datetime64": ProfilerDataTypes.DATETIME.value,
                           "datetime64[ns]": ProfilerDataTypes.DATETIME.value,
                           "datetime[64]": ProfilerDataTypes.DATETIME.value,
                           "timedelta[ns]": ProfilerDataTypes.DATETIME.value,
                           "category": ProfilerDataTypes.CATEGORICAL.value,
                           "list": ProfilerDataTypes.ARRAY.value,
                           "vector": ProfilerDataTypes.OBJECT.value,
                           "dict": ProfilerDataTypes.OBJECT.value}

    # short to internal 
    SHORT_DTYPES = {"str": "string",
                    "integer": "int",
                    "big": "bigint",
                    "long": "bigint",
                    "bool": "boolean", # TO DO: bool: (True, False), boolean: (True, False, Null)
                    "timestamp": "datetime"
                    }
                     
    @property
    def OPTIMUS_TO_INTERNALS(self):
        values = set(list(self.INTERNAL_TO_OPTIMUS.values()) + ProfilerDataTypes.list())
        
        _dict = {}
        for v in values:
            _list = [self.OPTIMUS_TO_INTERNAL[v]]
            _list += [t[0] for t in self.INTERNAL_TO_OPTIMUS.items() if t[1] == v]
            _dict.update({v: list(set(_list))})
            
        return _dict

    ANY_TYPES = ["object"]

    @property
    def INT_INTERNAL_TYPES(self):
        return [ProfilerDataTypes.INT.value] + [item[0] for item in self.INTERNAL_TO_OPTIMUS.items() if item[1] == ProfilerDataTypes.INT.value]

    @property
    def NUMERIC_INTERNAL_TYPES(self):
        types = [ProfilerDataTypes.INT.value, ProfilerDataTypes.FLOAT.value]
        return types + [item[0] for item in self.INTERNAL_TO_OPTIMUS.items() if item[1] in types]

    @property
    def STRING_INTERNAL_TYPES(self):
        return [ProfilerDataTypes.STRING.value] + [item[0] for item in self.INTERNAL_TO_OPTIMUS.items() if item[1] == ProfilerDataTypes.STRING.value]
    
    @property
    def DATETIME_INTERNAL_TYPES(self):
        return [ProfilerDataTypes.DATETIME.value] +\
               [item[0] for item in self.INTERNAL_TO_OPTIMUS.items() if item[1] == ProfilerDataTypes.DATETIME.value]

    @property
    def INT_TYPES(self):
        return self.ANY_TYPES + [ProfilerDataTypes.INT.value] +\
               [item[0] for item in self.INTERNAL_TO_OPTIMUS.items() if item[1] == ProfilerDataTypes.INT.value]

    @property
    def NUMERIC_TYPES(self):
        types = [ProfilerDataTypes.INT.value, ProfilerDataTypes.FLOAT.value]
        return self.ANY_TYPES + types +\
               [item[0] for item in self.INTERNAL_TO_OPTIMUS.items() if item[1] in types]

    @property
    def STRING_TYPES(self):
        return self.ANY_TYPES + [ProfilerDataTypes.STRING.value] +\
               [item[0] for item in self.INTERNAL_TO_OPTIMUS.items() if item[1] == ProfilerDataTypes.STRING.value]
    
    @property
    def DATETIME_TYPES(self):
        return self.ANY_TYPES + [ProfilerDataTypes.DATETIME.value] +\
               [item[0] for item in self.INTERNAL_TO_OPTIMUS.items() if item[1] == ProfilerDataTypes.DATETIME.value]

    @property
    def OBJECT_TYPES(self):
        types = [ProfilerDataTypes.OBJECT.value, ProfilerDataTypes.ARRAY.value]
        return self.ANY_TYPES + types +\
               [item[0] for item in self.INTERNAL_TO_OPTIMUS.items() if item[1] in types]

    COMPATIBLE_DTYPES = {}


LIMIT = 1000
LIMIT_TABLE = 10
NUM_PARTITIONS = 10
SAMPLE_NUMBER = 10000



import functools
import operator
from functools import reduce

from multipledispatch import dispatch
from pyspark.sql import DataFrame
from pyspark.sql import functions as F

# Helpers
from optimus.engines.base.dataframe.rows import DataFrameBaseRows
from optimus.engines.base.pandas.rows import PandasBaseRows
from optimus.engines.base.rows import BaseRows
from optimus.engines.spark.audf import filter_row_by_data_type as fbdt
from optimus.engines.base.dask.rows import DaskBaseRows
from optimus.engines.spark.create import Create
from optimus.helpers.check import is_spark_dataframe
from optimus.helpers.columns import parse_columns
from optimus.helpers.constants import Actions
from optimus.helpers.core import val_to_list, one_list_to_val
from optimus.helpers.functions_spark import append as append_df
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_list_of_tuples, is_list_of_str_or_int
# from optimus.infer_spark import is_list_of_spark_dataframes

from optimus.engines.base.meta import Meta
class Rows(DataFrameBaseRows, PandasBaseRows, BaseRows):
    pass
#
# class Rows(DaskBaseRows):
#
#     def __init__(self, df):
#         super(DaskBaseRows, self).__init__(df)
#
#     @staticmethod
#     def append(rows) -> DataFrame:
#         """
#         Append a row at the end of a dataframe
#         :param rows: List of tuples or dataframes to be appended
#         :return: Spark DataFrame
#         """
#         df = self
#
#         if is_list_of_tuples(rows):
#             columns = [str(i) for i in range(df.cols.count())]
#             if not is_list_of_tuples(rows):
#                 rows = [tuple(rows)]
#             new_row = Create().df(columns, rows)
#             df_result = df.union(new_row)
#
#         elif is_list_of_spark_dataframes(rows) or is_spark_dataframe(rows):
#             row = val_to_list(rows)
#             row.insert(0, df)
#             df_result = append_df(row, like="rows")
#         else:
#             RaiseIt.type_error(rows, ["list of tuples", "list of dataframes"])
#
#         df_result = df_result.meta.preserve(None, Actions.APPEND.value, df.cols.names())
#
#         return df_result
#
#     def select(self, condition):
#         """
#         Alias of Spark filter function. Return rows that match a expression
#         :param condition:
#         :return: Spark DataFrame
#         """
#         dfd = self.root.data
#         dfd = dfd.filter(condition)
#         df = self.root.new(dfd)
#         df.meta = Meta.action(df.meta, None, Actions.SORT_ROW.value, df.cols.names())
#
#         return odf
#
#     def count(self, compute=True) -> int:
#         """
#         Count dataframe rows
#         """
#
#         return self.root.data.count()
#
#     @staticmethod
#     def select(*args, **kwargs) -> DataFrame:
#         """
#         Alias of Spark filter function. Return rows that match a expression
#         :param args:
#         :param kwargs:
#         :return: Spark DataFrame
#         """
#         return self.root.data.filter(*args, **kwargs)
#
#     def _sort_multiple(self, dfd, col_sort, meta):
#         func = []
#         for cs in col_sort:
#             col_name = one_list_to_val(cs[0])
#             order = cs[1]
#
#             if order == "asc":
#                 sort_func = F.asc
#             elif order == "desc":
#                 sort_func = F.desc
#             else:
#                 RaiseIt.value_error(sort_func, ["asc", "desc"])
#
#             func.append(sort_func(col_name))
#             meta = Meta.action(meta, Actions.SORT_ROW.value, col_name)
#
#         dfd = dfd.sort(*func)
#         return dfd, meta
#
#
#     def _sort(self):
#         pass
#
#
#     def drop(self,where=None) -> DataFrame:
#         """
#         Drop a row depending on a dataframe expression
#         :param where: Expression used to drop the row
#         :return: Spark DataFrame
#         """
#         df = self.paren.data
#         df = df.where(~where)
#         df.meta = Meta.action(df.meta, None, Actions.DROP_ROW.value, df.cols.names())
#         return df
#
#     @staticmethod
#     def drop_by_dtypes(input_cols, data_type=None):
#         """
#         Drop rows by cell data type
#         :param input_cols: Column in which the filter is going to be applied
#         :param data_type: filter by string, integer, float or boolean
#         :return: Spark DataFrame
#         """
#         df = self
#         input_cols = parse_columns(df, input_cols)
#         df = df.rows.drop(fbdt(input_cols, data_type))
#         df.meta = Meta.action(df.meta, None, Actions.DROP_ROW.value, df.cols.names())
#         return df
#
#     @staticmethod
#     def drop_na(input_cols, how="any") -> DataFrame:
#         """
#         Removes rows with null values. You can choose to drop the row if 'all' values are nulls or if
#         'any' of the values is null.
#
#         :param input_cols:
#         :param how: ‘any’ or ‘all’. If ‘any’, drop a row if it contains any nulls. If ‘all’, drop a row only if all its
#         values are null. The default is 'all'.
#         :return: Returns a new DataFrame omitting rows with null values.
#         """
#         df = self
#         input_cols = parse_columns(self.root, input_cols)
#
#         df = df.dropna(how, subset=input_cols)
#         df.meta = Meta.action(df.meta, None, Actions.DROP_ROW.value, df.cols.names())
#         return df
#
#     def drop_duplicated(self, input_cols=None) -> DataFrame:
#         """
#         Drop duplicates values in a dataframe
#         :param input_cols: List of columns to make the comparison, this only  will consider this subset of columns,
#         :return: Return a new DataFrame with duplicate rows removed
#         """
#         # TODO:
#         #  add param
#         #  first : Drop duplicates except for the first occurrence.
#         #  last : Drop duplicates except for the last occurrence.
#         #  all: Drop all duplicates except for the last occurrence.
#
#         df = self.root
#         dfd = df.data
#
#         input_cols = parse_columns(df, input_cols)
#
#         dfd = dfd.drop_duplicates(subset=input_cols)
#         meta = Meta.action(df.meta, None, Actions.DROP_ROW.value, df.cols.names())
#
#         return self.root.new(dfd, meta=meta)
#
#     @staticmethod
#     def drop_first() -> DataFrame:
#         """
#         Remove first row in a dataframe
#         :return: Spark DataFrame
#         """
#         df = self
#         df = df.zipWithIndex().filter(lambda tup: tup[1] > 0).map(lambda tup: tup[0])
#         df.meta = Meta.action(df.meta, None, Actions.DROP_ROW.value, df.cols.names())
#         return df
#
#
#     def limit(self, count):
#         """
#         Limit the number of rows
#         :param count:
#         :return:
#         """
#         dfd = self.root.data
#         return self.root.new(dfd.limit(count))
#
#     # TODO: Merge with select
#     @staticmethod
#     def is_in(input_cols, values) -> DataFrame:
#         """
#         Filter rows which columns match a specific value
#         :return: Spark DataFrame
#         """
#         df = self
#
#         # Ensure that we have a list
#         values = val_to_list(values)
#
#         # Create column/value expression
#         column_expr = [(F.col(input_cols) == v) for v in values]
#
#         # Concat expression with and logical or
#         expr = reduce(lambda a, b: a | b, column_expr)
#         df = df.rows.select(expr)
#         df.meta = Meta.action(df.meta, None, Actions.DROP_ROW.value, input_cols)
#         return df
#
#     @staticmethod
#     def unnest(input_cols) -> DataFrame:
#         """
#         Convert a list in a rows to multiple rows with the list values
#         :param input_cols:
#         :return:
#         """
#         input_cols = parse_columns(self.root, input_cols)[0]
#         df = self.root
#         dfd = df.data
#         dfd = dfd.withColumn(input_cols, F.explode(input_cols))
#         meta = Meta.set(df.meta, value=df.meta.preserve(None, Actions.DROP_ROW.value, input_cols).get())
#         return df.new(dfd, meta=meta)
#
#     def approx_count(self, timeout=1000, confidence=0.90) -> DataFrame:
#         """
#         Return aprox rows count
#         :param timeout:
#         :param confidence:
#         :return:
#         """
#         dfd = self.root.data
#         return dfd.rdd.countApprox(timeout, confidence)



import cudf
import dask_cudf

from optimus.engines.base.cudf.functions import CUDFBaseFunctions
from optimus.engines.base.dask.functions import DaskBaseFunctions

class DaskCUDFFunctions(CUDFBaseFunctions, DaskBaseFunctions):

    _partition_engine = cudf

    @staticmethod
    def from_dataframe(dfd):
        dask_cudf.from_cudf(dfd)

    @staticmethod
    def dask_to_compatible(dfd):
        from optimus.helpers.converter import dask_dataframe_to_dask_cudf
        return dask_dataframe_to_dask_cudf(dfd)

    def kurtosis(self, series):
        return series.map_partitions(lambda _series: _series.kurtosis())

    def skew(self, series):
        return series.map_partitions(lambda _series: _series.skew())

    def sqrt(self, series):
        return series.map_partitions(lambda _series: _series.sqrt())

    def exp(self, series):
        return series.map_partitions(lambda _series: _series.exp())

    def ln(self, series):
        return series.map_partitions(lambda _series: _series.log())

    def radians(self, series):
        return cudf.radians(series.to_float())

    def degrees(self, series):
        return cudf.degrees(series.to_float())

    def log(self, series, base=10):
        return series.map_partitions(lambda _series: _series.log()) / cudf.log(base)

    def ceil(self, series):
        return series.map_partitions(lambda _series: _series.ceil())

    def floor(self, series):
        return series.map_partitions(lambda _series: _series.floor())

    def sin(self, series):
        return series.map_partitions(lambda _series: _series.sin())

    def cos(self, series):
        return series.map_partitions(lambda _series: _series.cos())

    def tan(self, series):
        return series.map_partitions(lambda _series: _series.tan())

    def asin(self, series):
        return series.map_partitions(lambda _series: _series.asin())

    def acos(self, series):
        return series.map_partitions(lambda _series: _series.acos())

    def atan(self, series):
        return series.map_partitions(lambda _series: _series.atan())

    def sinh(self, series):
        return 1 / 2 * (self.exp() - self.exp())

    def cosh(self, series):
        return 1 / 2 * (self.exp() + self.exp())

    def tanh(self):
        return self.sinh() / self.cosh()

    def asinh(self):
        return 1 / self.sinh()

    def acosh(self):
        return 1 / self.cosh()

    def atanh(self):
        return 1 / self.tanh()

    def cut(self, series, bins, labels):
        raise NotImplementedError

    def normalize_chars(self, series):
        # str.decode return a float column. We are forcing to return a string again
        return self.to_string_accessor(series).normalize_characters()

    def remove_special_chars(self, series):
        # See https://github.com/rapidsai/cudf/issues/5520
        return self.to_string_accessor(series).replace_non_alphanumns(replacement_char='')

    def date_format(self, series, current_format=None, output_format=None):
        return cudf.to_datetime(series).astype('str', format=output_format)

    def time_between(self, date_format=None):
        raise NotImplementedError("Not implemented yet see https://github.com/rapidsai/cudf/issues/1041")
        # return cudf.to_datetime(series).astype('str', format=date_format) - datetime.now().date()




import numpy as np
import databricks.koalas as ks

from optimus.engines.base.commons.functions import word_tokenize

from optimus.engines.base.dataframe.functions import DataFrameBaseFunctions
from optimus.engines.base.pandas.functions import PandasBaseFunctions


# These function can return a Column Expression or a list of columns expression
# Must return None if the data type can not be handle


class SparkFunctions(PandasBaseFunctions, DataFrameBaseFunctions):

    _engine = ks

    @staticmethod
    def dask_to_compatible(dfd):
        from optimus.helpers.converter import dask_dataframe_to_pandas
        return ks.from_pandas(dask_dataframe_to_pandas(dfd))

    @staticmethod
    def df_concat(df_list):
        return ks.concat(df_list, axis=0, ignore_index=True)

    def word_tokenize(self, series):
        return self.to_string(series).map(word_tokenize, na_action=None)

    def count_zeros(self, series, *args):
        return int((self.to_float(series).values == 0).sum())

    def kurtosis(self, series):
        return self.to_float(series).kurtosis()

    def skew(self, series):
        return self.to_float(series).skew()

    def exp(self, series):
        return np.exp(self.to_float(series))

    def sqrt(self, series):
        return np.sqrt(self.to_float(series))

    def reciprocal(self, series):
        return np.reciprocal(self.to_float(series))

    def radians(self, series):
        return np.radians(self.to_float(series))

    def degrees(self, series):
        return np.degrees(self.to_float(series))

    def ln(self, series):
        return np.log(self.to_float(series))

    def log(self, series, base=10):
        return np.log(self.to_float(series)) / np.log(base)

    def sin(self, series):
        return np.sin(self.to_float(series))

    def cos(self, series):
        return np.cos(self.to_float(series))

    def tan(self, series):
        return np.tan(self.to_float(series))

    def asin(self, series):
        return np.arcsin(self.to_float(series))

    def acos(self, series):
        return np.arccos(self.to_float(series))

    def atan(self, series):
        return np.arctan(self.to_float(series))

    def sinh(self, series):
        return np.arcsinh(self.to_float(series))

    def cosh(self, series):
        return np.cosh(self.to_float(series))

    def tanh(self, series):
        return np.tanh(self.to_float(series))

    def asinh(self, series):
        return np.arcsinh(self.to_float(series))

    def acosh(self, series):
        return np.arccosh(self.to_float(series))

    def atanh(self, series):
        return np.arctanh(self.to_float(series))

    def floor(self, series):
        return np.floor(self.to_float(series))

    def ceil(self, series):
        return np.ceil(self.to_float(series))

    def normalize_chars(self, series):
        return series.str.normalize("NFKD").str.encode('ascii', errors='ignore').str.decode('utf8')

    def format_date(self, series, current_format=None, output_format=None):
        return ks.to_datetime(series, format=current_format,
                              errors="coerce").dt.strftime(output_format).reset_index(drop=True)

    def time_between(self, series, value=None, date_format=None):

        value_date_format = date_format

        if is_list_or_tuple(date_format) and len(date_format) == 2:
            date_format, value_date_format = date_format

        if is_list_or_tuple(value) and len(value) == 2:
            value, value_date_format = value

        date = pd.to_datetime(series, format=date_format, errors="coerce")
        value = pd.Timestamp.now() if value is None else pd.to_datetime(value, format=value_date_format, errors="coerce")

        return (value - date)


from pyspark.ml.linalg import VectorUDT
from pyspark.sql.types import StringType, BooleanType, IntegerType, ArrayType, FloatType, DoubleType, StructType, \
    DateType, LongType, ByteType, ShortType, TimestampType, BinaryType, NullType

from optimus.engines.base.constants import BaseConstants

# Python to PySpark reference
#
# type(None): NullType,
# bool: BooleanType,
# int: LongType,
# float: DoubleType,
# str: StringType,
# bytearray: BinaryType,
# decimal.Decimal: DecimalType,
# datetime.date: DateType,
# datetime.datetime: TimestampType,
# datetime.time: TimestampType,


DTYPES_DICT_OBJECTS = \
    {"string": StringType(), "int": IntegerType(), "float": FloatType(),
     "double": DoubleType(), "boolean": BooleanType(), "struct": StructType(), "array": ArrayType(StringType()),
     "bigint": LongType(), "date": DateType(), "byte": ByteType(), "short": ShortType(),
     "datetime": TimestampType(), "binary": BinaryType(), "null": NullType()
     }

SHORT_DTYPES = {"string": "string",
                "str": "string",
                "integer": "int",
                "int": "int",
                "bigint": "bigint",
                "big": "bigint",
                "long": "bigint",
                "float": "float",
                "double": "double",
                "bool": "boolean",
                "boolean": "boolean",
                "struct": "struct",
                "array": "array",
                "date": "date",
                "datetime": "datetime",
                "byte": "byte",
                "short": "short",
                "binary": "binary",
                "null": "null",
                "vector": "vector",
                "timestamp": "datetime"
                }

SPARK_VERSION = "2.4.1"
HADOOP_VERSION = "2.7"
SPARK_FILE = "spark-{SPARK_VERSION}-bin-hadoop{HADOOP_VERSION}.tgz".format(SPARK_VERSION=SPARK_VERSION,
                                                                           HADOOP_VERSION=HADOOP_VERSION)
SPARK_URL = "https://archive.apache.org/dist/spark/spark-{SPARK_VERSION}//{SPARK_FILE}".format(
    SPARK_VERSION=SPARK_VERSION, SPARK_FILE=SPARK_FILE)
SPARK_PATH_COLAB = "/content/spark-{SPARK_VERSION}-bin-hadoop{HADOOP_VERSION}".format(SPARK_VERSION=SPARK_VERSION,
                                                                                      HADOOP_VERSION=HADOOP_VERSION)
STARTING_SPARK = "Starting or getting SparkSession and SparkContext..."


class Constants(BaseConstants):
    SHORT_DTYPES = SHORT_DTYPES

    DTYPES_DICT = {"string": StringType, "int": IntegerType, "float": FloatType,
                   "double": DoubleType, "boolean": BooleanType, "struct": StructType, "array": ArrayType,
                   "bigint": LongType, "date": DateType, "byte": ByteType, "short": ShortType,
                   "datetime": TimestampType, "binary": BinaryType, "null": NullType, "vector": VectorUDT
                   }

    DTYPES_DICT_OBJECTS = DTYPES_DICT_OBJECTS

    NUMERIC_TYPES = ["byte", "short", "big", "int", "double", "float"]
    NOT_ARRAY_TYPES = ["byte", "short", "big", "int", "double", "float", "string", "date", "bool"]
    STRING_TYPES = ["str"]
    ARRAY_TYPES = ["array"]

    DTYPES_TO_INFERRED = {"int": ["smallint", "tinyint", "bigint", "int"], "float": ["float", "double"],
                          "string": ["string"], "date": ["date", "timestamp"], "boolean": ["boolean"],
                          "binary": ["binary"],
                          "array": ["array"], "object": ["object"], "null": ["null"], "missing": ["missing"]}



import re

import pandas as pd
from ibis.expr.types import TableExpr
from sklearn import preprocessing

from optimus.engines.base.commons.functions import string_to_index, index_to_string
from optimus.engines.base.dataframe.columns import DataFrameBaseColumns
from optimus.helpers.columns import parse_columns, prepare_columns
from optimus.helpers.constants import Actions
from optimus.helpers.converter import format_dict
from optimus.helpers.core import val_to_list
from optimus.infer import is_str, is_tuple, is_dict

DataFrame = TableExpr


class Cols(DataFrameBaseColumns):
    def __init__(self, df):
        super().__init__(df)

    def _names(self):
        return list(self.root.data.columns)

    def append(self, dfs):
        """

        :param dfs:
        :return:
        """

        dfd = self.root.data
        dfd = pd.concat([dfs.data.reset_index(drop=True), dfd.reset_index(drop=True)], axis=1)
        return self.root.new(dfd)

    def data_type(self, columns="*"):
        df = self.root
        columns = parse_columns(df, columns)
        return dict(df.data[columns].schema().items())

    def agg_exprs(self, columns, funcs, *args, compute=True, tidy=True):
        df = self.root
        columns = parse_columns(df, columns)

        funcs = val_to_list(funcs)
        all_funcs = []

        for col_name in columns:
            for func in funcs:
                all_funcs.append({func.__name__: {col_name: self.exec_agg(func(df.data[col_name], *args))}})

        result = {}
        for i in all_funcs:
            for x, y in i.items():
                result.setdefault(x, {}).update(y)

        return format_dict(result, tidy)

    @staticmethod
    def exec_agg(exprs, compute=None):
        """
        Execute an aggregation
        :param exprs: Aggreagtion function to process
        :return:
        """
        if getattr(exprs, "execute", None) and compute:
            result = exprs.execute()
        else:
            result = exprs
            
        while isinstance(result, (list, tuple)) and len(result) == 1:
            result = result[0]

        return result

    @staticmethod
    def to_timestamp(input_cols, date_format=None, output_cols=None):
        pass

    @staticmethod
    def astype(*args, **kwargs):
        pass

    def apply(self, input_cols, func=None, func_return_type=None, args=None, func_type=None, when=None,
              filter_col_by_dtypes=None, output_cols=None, skip_output_cols_processing=False,
              meta_action=Actions.APPLY_COLS.value, mode="partitioned", set_index=False, default=None, **kwargs):
        columns = prepare_columns(self.root, input_cols, output_cols, filter_by_column_types=filter_col_by_dtypes,
                                  accepts_missing_cols=True, default=default)
        kw_columns = {}
        if args is None:
            args = (None,)
        elif not is_tuple(args, ):
            args = (args,)

        for input_col, output_col in columns:
            # print("args",args)
            kw_columns.update({output_col: func(self.root.data[input_col], *args)})
        return self.root.new(self.root.data.mutate(**kw_columns))

    def find(self, columns, sub, ignore_case=False):
        """
        Find the start and end position for a char or substring
        :param columns:
        :param ignore_case:
        :param sub:
        :return:
        """
        df = self
        columns = parse_columns(df, columns)
        sub = val_to_list(sub)

        def get_match_positions(_value, _separator):

            result = None
            if is_str(_value):
                # Using re.IGNORECASE in finditer not seems to work
                if ignore_case is True:
                    _separator = _separator + [s.lower() for s in _separator]
                regex = re.compile('|'.join(_separator))

                length = [[match.start(), match.end()] for match in
                          regex.finditer(_value)]
                result = length if len(length) > 0 else None
            return result

        for col_name in columns:
            # Categorical columns can not handle a list inside a list as return for example [[1,2],[6,7]].
            # That could happened if we try to split a categorical column
            # df[col_name] = df[col_name].astype("object")
            df[col_name + "__match_positions__"] = df[col_name].astype("object").apply(get_match_positions,
                                                                                       args=(sub,))
        return df

    @staticmethod
    def correlation(input_cols, method="pearson", output="json"):
        pass

    def qcut(self, columns, quantiles, handle_invalid="skip"):
        pass

    def string_to_index(self, cols=None, output_cols=None):
        df = self.df
        le = df.le or preprocessing.LabelEncoder()
        df = string_to_index(df, cols, output_cols, le)

        return df

    def index_to_string(self, cols=None, output_cols=None):
        df = self.df
        le = df.le or preprocessing.LabelEncoder()
        df = index_to_string(df, cols, output_cols, le)

        return df



import pandas as pd

DataFrame = pd.DataFrame
from optimus.helpers.logger import logger
from optimus.engines.base.io.save import BaseSave


def save(self: DataFrame):
    class Save(BaseSave):
        @staticmethod
        def json(path, *args, **kwargs):
            df = self
            try:
                df.to_json(path, *args, **kwargs)

            except IOError as e:
                logger.print(e)
                raise

        @staticmethod
        def csv(path, mode="w", **kwargs):

            try:
                df = self
                # columns = parse_columns(self, "*",
                #                         filter_by_column_types=["date", "array", "vector", "binary", "null"])
                # df = df.cols.cast(columns, "str").repartition(num_partitions)

                # Dask reference
                # https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.to_csv
                # df.to_csv(filename=path)
                df.to_csv(path, index=False, mode=mode)

            except IOError as error:
                logger.print(error)
                raise

        @staticmethod
        def parquet(path, mode="overwrite", num_partitions=1):

            # This character are invalid as column names by parquet
            invalid_character = [" ", ",", ";", "{", "}", "(", ")", "\n", "\t", "="]

            def func(col_name):
                for i in invalid_character:
                    col_name = col_name.replace(i, "_")
                return col_name

            df = self.cols.rename(func)

            try:
                df.to_parquet(path, num_partitions=num_partitions)
            except IOError as e:
                logger.print(e)
                raise

        @staticmethod
        def avro(path):
            raise NotImplementedError('Not implemented yet')

        @staticmethod
        def rabbit_mq(host, exchange_name=None, queue_name=None, routing_key=None, parallelism=None):
            raise NotImplementedError('Not implemented yet')

    return Save()


DataFrame.save = property(save)



from enum import Enum

from optimus.engines.base.io.drivers.abstract_driver import AbstractDriver
from optimus.engines.base.io.properties import DriverProperties
from optimus.helpers.functions import singleton


@singleton
class SQLServerDriver(AbstractDriver):
    """SQLServer Database"""

    def properties(self) -> Enum:
        return DriverProperties.SQLSERVER

    def uri(self, *args, **kwargs) -> str:
        if kwargs["sso"]:
            return f"""{kwargs["driver"]}://{kwargs["host"]}:{kwargs["port"]}/{kwargs["database"]}?driver=ODBC+Driver+17+for+SQL+Server&Trusted_Connection=yes"""
        else:
            return f"""{kwargs["driver"]}://{kwargs["user"]}:{kwargs["password"]}@{kwargs["host"]}:{kwargs["port"]}/{kwargs["database"]}?driver=ODBC+Driver+17+for+SQL+Server"""

    def url(self, *args, **kwargs) -> str:
        return f"""jdbc:{kwargs["driver"]}://{kwargs["host"]}:{kwargs["port"]};databaseName={kwargs["database"]}"""

    def table_names_query(self, *args, **kwargs):
        return "SELECT * FROM INFORMATION_SCHEMA.TABLES"

    def table_name_query(self, *args, **kwargs) -> str:
        return "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES"

    def count_query(self, *args, **kwargs) -> str:
        return "SELECT COUNT(*) as COUNT FROM " + kwargs["db_table"]

    def primary_key_query(self, *args, **kwargs) -> str:
        pass

    def min_max_query(self, *args, **kwargs) -> str:
        return f"""SELECT min({kwargs["partition_column"]}) AS min, max({kwargs["partition_column"]}) AS max FROM {
        kwargs["table_name"]} """



HEADER = '''<!doctype html>
<html class="no-js" lang="">

<head>
  <meta charset="utf-8">
  <meta http-equiv="x-ua-compatible" content="ie=edge">
  <title></title>
  <meta name="description" content="">
  <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">

  <link rel="manifest" href="site.webmanifest">
  <link rel="apple-touch-icon" href="icon.png">
  <!-- Place favicon.ico in the root directory -->

</head>

<body>'''

FOOTER = '''</body></html>'''



import ntpath

import dask.bag as db
import pandas as pd
from dask import dataframe as dd

from optimus.engines.base.basedataframe import BaseDataFrame
from optimus.engines.base.io.load import BaseLoad
from optimus.engines.base.meta import Meta
from optimus.engines.dask_cudf.dataframe import DaskCUDFDataFrame
from optimus.helpers.functions import prepare_path, unquote_path
from optimus.helpers.logger import logger


class Load(BaseLoad):

    def json(self, path, multiline=False, storage_options=None, conn=None, *args, **kwargs):
        """
        Return a dask dataframe from a json file.
        :param path: path or location of the file.
        :param multiline:

        :return:
        """

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "json")[0]

        try:
            import dask_cudf
            df = dask_cudf.read_json(path, lines=multiline, storage_options=storage_options, *args, **kwargs)
            df = DaskCUDFDataFrame(df, op=self.op)
            df.meta = Meta.set(df.meta, "file_name", file_name)

        except IOError as error:
            logger.print(error)
            raise
        return df

    def csv(self, filepath_or_buffer, sep=',', header=True, infer_schema=True, encoding="utf-8", null_value="None", n_rows=-1, cache=False,
            quoting=0, lineterminator=None, on_bad_lines=False, engine="c", keep_default_na=True, na_filter=True,
            storage_options=None, conn=None, *args, **kwargs):


        filepath_or_buffer = unquote_path(filepath_or_buffer)

        if conn is not None:
            filepath_or_buffer = conn.path(filepath_or_buffer)
            storage_options = conn.storage_options

        remove_param = "chunk_size"
        if kwargs.get(remove_param):
            # This is handle in this way to preserve compatibility with others dataframe technologies.
            logger.print(f"{remove_param} is not supported. Used to preserve compatibility with Optimus Pandas")
            kwargs.pop(remove_param)

        try:
            import dask_cudf
            if engine == "python":

                # na_filter=na_filter, on_bad_lines and low_memory are not support by pandas engine
                dcdf = dask_cudf.read_csv(filepath_or_buffer, sep=sep, header=0 if header else None, encoding=encoding,
                                          quoting=quoting, keep_default_na=True, na_values=None, engine=engine,
                                          storage_options=storage_options, on_bad_lines='skip', *args, **kwargs)

            elif engine == "c":
                dcdf = dask_cudf.read_csv(filepath_or_buffer, sep=sep, header=0 if header else None, encoding=encoding,
                                          quoting=quoting, on_bad_lines=on_bad_lines,
                                          keep_default_na=True, na_values=None, engine=engine, na_filter=na_filter,
                                          storage_options=storage_options, low_memory=False, *args, **kwargs)

            if n_rows > -1:
                dcdf = dask_cudf.from_cudf(dcdf.head(n=n_rows), npartitions=1).reset_index(drop=True)

            df = DaskCUDFDataFrame(dcdf, op=self.op)
            df.meta = Meta.set(df.meta, "file_name", filepath_or_buffer)
            df.meta = Meta.set(df.meta, "name", ntpath.basename(filepath_or_buffer))
        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def parquet(path, columns=None, storage_options=None, conn=None, *args, **kwargs):

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "parquet")

        try:
            import dask_cudf
            df = dask_cudf.read_parquet(path, columns=columns, storage_options=storage_options, *args, **kwargs)

            df.meta = Meta.set(df.meta, "file_name", file_name)

        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def avro(path, storage_options=None, conn=None, *args, **kwargs):


        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "avro")

        try:
            df = db.read_avro(path, storage_options=storage_options, *args, **kwargs).to_dataframe()
            df.meta = Meta.set(df.meta, "file_name", file_name)

        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def excel(path, sheet_name=0, storage_options=None, conn=None, *args, **kwargs):

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "xls")

        try:
            pdf = pd.read_excel(file, sheet_name=sheet_name, storage_options=storage_options, *args, **kwargs)

            # Parse object column data type to string to ensure that Spark can handle it. With this we try to reduce
            # exception when Spark try to infer the column data type
            col_names = list(pdf.select_dtypes(include=['object']))

            column_dtype = {}
            for col in col_names:
                column_dtype[col] = str

            # Convert object columns to string
            pdf = pdf.astype(column_dtype)

            # Create data frame
            df = dd.from_pandas(pdf, npartitions=3)
            df.meta = Meta.set(df.meta, "file_name", ntpath.basename(file_name))
        except IOError as error:
            logger.print(error)
            raise

        return df



import re
import inspect
from inspect import signature
from pprint import pformat
from typing import List, Union, Optional
from multipledispatch.dispatcher import MethodDispatcher

from optimus.engines.base.basedataframe import BaseDataFrame
from optimus.infer import is_list, is_str
from optimus.helpers.types import *
from optimus.helpers.core import val_to_list, one_list_to_val

from .prepare import prepare

from optimus.engines.base.engine import BaseEngine
from optimus.engines.base.stringclustering import Clusters

from optimus.engines.base.create import BaseCreate
from optimus.engines.base.io.load import BaseLoad
from optimus.engines.base.io.save import BaseSave
from optimus.engines.base.io.connect import Connect

from optimus.engines.base.columns import BaseColumns
from optimus.engines.base.rows import BaseRows
from optimus.engines.base.set import BaseSet
from optimus.engines.base.mask import Mask
from optimus.engines.base.ml.models import BaseML
from optimus.plots.plots import Plot
from optimus.outliers.outliers import Outliers
from optimus.engines.base.profile import BaseProfile

engine_accessors = {
    "create": BaseCreate,
    "load": BaseLoad,
    "save": BaseSave,
    "connect": Connect
}

dataframe_accessors = {
    "cols": BaseColumns,
    "rows": BaseRows,
    "set": BaseSet,
    "mask": Mask,
    "ml": BaseML,
    "plot": Plot,
    "outliers": Outliers,
    "profile": BaseProfile
}

accessors = {**engine_accessors, **dataframe_accessors}

def _create_new_variable(base_name: str, names: List[str]):
    while base_name in names:
        base_name = _increment_variable_name(base_name)
        
    return base_name
    
def _increment_variable_name(variable_name: str):
    match = re.search(r'\d+$', variable_name)
    if match:
        variable_name = variable_name[0:match.start()] + str(int(variable_name[match.start():match.endpos])+1)
    else:
        variable_name = variable_name + "2"
    return variable_name

def _arguments(args: dict, args_properties: Optional[dict]=None):
    
    args_list = []
    
    # It doesn't find properties or allows kwargs
    if not args_properties or "kwargs" in args_properties:
        args_list = list(args.keys())
        for key in ["source", "target", "operation", "operation_options"]:
            if key in args_list:
                args_list.remove(key)
    
    # Properties from declaration
    else:
        args_list = args_properties.keys()

    # Formating
    for arg in args_list:
        # Variable names of list of variable names
        if arg in args:
            if arg in args_properties and is_any_optimus_type(args_properties[arg].get("type", None)):
                if is_list(args[arg]):
                    args[arg] = f"[{', '.join(args[arg])}]"
                else:
                    args[arg] = args[arg]
            # Native types
            elif is_str(arg) and arg in args_properties and args_properties[arg].get("type", None) == dict:
                pass
            else:
                args[arg] = pformat(args[arg])

        
    return ", ".join([f'{arg}={args[arg]}' for arg in args_list if arg in args])

def _generate_code_target(body: dict, properties: dict, target: str):

    arguments = _arguments(body, properties["arguments"])
    
    code = ''

    if target:
        code += f'{target} = '
    
    if body.get("source", None):
        code += f'{body["source"]}.'

    code += f'{body["operation"]}({arguments})'
    
    if target:
        return code, [target]
    else:
        return code, []

def _generate_code_dataframe_transformation(body: dict, properties: dict, variables):

    target = body.get("target")

    if target is None:
        target = body["source"]

    options = body.get("operation_options", None)
    if options:
      if options.get("creates_new", False):
        target = available_variable("df", variables)
    return _generate_code_target(body, properties, target)

def _generate_code_dataframe_mask(body, properties, variables):
    target = body.get("target", available_variable("mask", variables))
    return _generate_code_target(body, properties, target)

def _generate_code_dataframe_clusters(body, properties, variables):    
    target = body.get("target", available_variable("clusters", variables))
    return _generate_code_target(body, properties, target)

def _generate_code_output(body, properties, variables):    
    target = body.get("target", "result")
    return _generate_code_target(body, properties, target)

def _generate_code_engine_dataframe(body, properties, variables):    
    target = body.get("target", available_variable("df", variables))
    return _generate_code_target(body, properties, target)

def _generate_code_engine_connection(body, properties, variables):    
    target = body.get("target", available_variable("conn", variables))
    return _generate_code_target(body, properties, target)

def _generate_code_engine(body, properties, variables):    
    target = body.get("target", available_variable("op", variables))
    return _generate_code_target(body, properties, target)

def _get_generator(func_properties, method_root_type):

    if method_root_type == "dataframe":
    
        # If the method returns a dataframe, it's a transformation
        if 'DataFrameType' == func_properties["return_annotation"]:
            return _generate_code_dataframe_transformation
        # TO-DO: should mask functions be treated as transformations? (MaskDataFrameType)
        if 'MaskDataFrameType' == func_properties["return_annotation"]:
            return _generate_code_dataframe_mask
        # If the method returns anything else, it's an output
        elif 'ClustersType' == func_properties["return_annotation"]:
            return _generate_code_dataframe_clusters
        else:
            return _generate_code_output
    
    elif method_root_type == "engine":
    
        if 'DataFrameType' == func_properties["return_annotation"]:
            return _generate_code_engine_dataframe
        elif 'ConnectionType' == func_properties["return_annotation"]:
            return _generate_code_engine_connection
        else:
            return _generate_code_output
        
    elif method_root_type == "optimus":
        
        return _generate_code_engine
    
    elif method_root_type == "clusters":
    
        return _generate_code_output
        
def _get_method_root_type(accessor):
    if accessor in engine_accessors:
        return "engine"
    if accessor in dataframe_accessors:
        return "dataframe"

def _init_methods(engine):

    method = ""

    if engine == "pandas":
        from optimus.engines.pandas.engine import PandasEngine
        method = PandasEngine
    if engine == "vaex":
        from optimus.engines.vaex.engine import VaexEngine
        method = VaexEngine
    if engine == "spark":
        from optimus.engines.spark.engine import SparkEngine
        method = SparkEngine
    if engine == "dask":
        from optimus.engines.dask.engine import DaskEngine
        method = DaskEngine
    if engine == "ibis":
        from optimus.engines.ibis.engine import IbisEngine
        method = IbisEngine
    if engine == "cudf":
        from optimus.engines.cudf.engine import CUDFEngine
        method = CUDFEngine
    if engine == "dask_cudf":
        from optimus.engines.dask_cudf.engine import DaskCUDFEngine
        method = DaskCUDFEngine
        
    return method

def _generate_code(body=None, variables=[], **kwargs):
    
    if not body:
        body = kwargs

    operation = body["operation"].split(".")

    method = None
    method_root_type = None
    
    if operation[0] == "Optimus":
        method = _init_methods(body["engine"])
        method_root_type = "optimus"
        operation = []
    elif operation[0] in accessors:
        method = accessors[operation[0]]
        method_root_type = _get_method_root_type(operation[0])
        operation = operation[1:]
    elif getattr(BaseDataFrame, operation[0], None):
        method = BaseDataFrame
        method_root_type = "dataframe"
    elif getattr(BaseEngine, operation[0], None):
        method = BaseEngine
        method_root_type = "engine"
    elif getattr(Clusters, operation[0], None):
        method = Clusters
        method_root_type = "clusters"

    for item in operation:
        method = getattr(method, item)

    properties = method_properties(method, method_root_type)

    code, updated = properties["generator"](body, properties, variables)
    
    return code, updated

def optimus_variables():
    from optimus.helpers.functions import list_engines, list_dataframes, list_clusters, list_connections
    return [ *list_engines(), *list_dataframes(), *list_clusters(), *list_connections() ]

def available_variable(name: str, variables: dict):
    return _create_new_variable(name, [*variables, *optimus_variables()])


def method_properties(func: Union[callable, MethodDispatcher], method_root_type):
    
    try:
        fp = signature(func)
        func_properties = {"parameters": fp.parameters, "return_annotation": fp.return_annotation}
    except ValueError as e:
        if getattr(func, "funcs"):
            func_properties = {"parameters": {}, "return_annotation": None}
            for f in func.funcs.values():
                fp = signature(f)
                func_properties["parameters"].update(fp.parameters)
                if fp.return_annotation:
                    func_properties["return_annotation"] = fp.return_annotation
        else:
            raise e

    if list(func_properties["parameters"].keys()) == ['root'] and getattr(func, "__call__"):
        fp = signature(func.__call__)
        func_properties = {"parameters": fp.parameters, "return_annotation": fp.return_annotation}

    arguments_list = list(func_properties["parameters"].items())
    arguments = {}
    for key, arg in arguments_list:
        
        if arg.name in ['self', 'cls']:
            continue
        
        arguments[arg.name] = {}
        
        if arg.annotation is not inspect._empty:
            arguments[arg.name].update({"type": arg.annotation})
            
        if arg.default is not inspect._empty:
            arguments[arg.name].update({"value": arg.default})
                        
    return {
        "arguments": arguments, 
        "returns": func_properties["return_annotation"],
        "generator": _get_generator(func_properties, method_root_type)
    }

def generate_code(body: Optional[dict]=None, variables: List[str]=[], get_updated: bool=False, **kwargs):

    if not body:
        body = kwargs

    body = val_to_list(body)

    updated = []

    code = []

    for operation in body:
        
        operation = prepare(operation)
        operation_code, operation_updated = _generate_code(operation, [*updated, *variables])
        updated.extend(operation_updated)
        code.append(operation_code)

    if get_updated:
        return "\n".join(code), one_list_to_val(updated)
    else:
        return "\n".join(code)
    



from imgkit import imgkit
from pyspark.ml.feature import SQLTransformer
from pyspark.serializers import AutoBatchedSerializer, PickleSerializer

from optimus.helpers.types import *
from optimus.engines.base.basedataframe import BaseDataFrame
from optimus.engines.pandas.dataframe import PandasDataFrame
from optimus.helpers.core import val_to_list
from optimus.helpers.functions import random_int, absolute_path
from optimus.helpers.output import print_html
from optimus.helpers.raiseit import RaiseIt


class SparkDataFrame(BaseDataFrame):
    def _base_to_dfd(self, pdf, n_partitions) -> 'InternalDataFrameType':
        pass

    def visualize(self):
        pass

    def graph(self) -> dict:
        pass

    @property
    def rows(self):
        from optimus.engines.spark.rows import Rows
        return Rows(self)

    @property
    def cols(self):
        from optimus.engines.spark.columns import Cols
        return Cols(self)

    @property
    def save(self):
        from optimus.engines.spark.io.save import Save
        return Save(self)

    @property
    def constants(self):
        from optimus.engines.spark.constants import Constants
        return Constants()

    @property
    def functions(self):
        from optimus.engines.spark.functions import SparkFunctions
        return SparkFunctions(self)

    def _buffer_window(self, input_cols, lower_bound, upper_bound):
        return self.__class__(self.data[input_cols][lower_bound: upper_bound], op=self.op, label_encoder=self.le)

    def cache(self):
        df = self.data
        return self.new(df.cache(), meta=self.meta)

    # @staticmethod
    # def to_json():
    #     """
    #     Return a json from a Spark Dataframe
    #
    #     :return:
    #     """
    #     return json.dumps(collect_as_dict(self), ensure_ascii=False, default=json_converter)

    # def to_dict(self, index=True):
    #     """
    #     Return a Python object from a Spark Dataframe
    #     :return:
    #     """
    #     return collect_as_dict(self.parent.data)

    @staticmethod
    def export():
        """
        Helper function to export all the dataframe in text format. Aimed to be used in test functions
        :return:
        """
        dict_result = {}
        value = self.collect()
        schema = []
        for col_names in self.cols.names():
            name = col_names

            data_type = self.cols.schema_data_type(col_names)
            if isinstance(data_type, ArrayType):
                data_type = "ArrayType(" + str(data_type.elementType) + "()," + str(data_type.containsNull) + ")"
            else:
                data_type = str(data_type) + "()"

            nullable = self.schema[col_names].nullable

            schema.append(
                "('{name}', {dataType}, {nullable})".format(name=name, dataType=data_type, nullable=nullable))
        schema = ",".join(schema)
        schema = "[" + schema + "]"

        # if there is only an element in the dict just return the value
        if len(dict_result) == 1:
            dict_result = next(iter(dict_result.values()))
        else:
            dict_result = [tuple(v.asDict().values()) for v in value]

        def func(path, _value):
            try:
                if math.isnan(_value):
                    r = None
                else:
                    r = _value
            except TypeError:
                r = _value
            return r

        dict_result = traverse(dict_result, None, func)

        return "{schema}, {dict_result}".format(schema=schema, dict_result=dict_result)

    @staticmethod
    def sample(n=10, random=False):
        """
        Return a n number of sample from a dataFrame
        :param n: Number of samples
        :param random: if true get a semi random sample
        :return:
        """
        if random is True:
            seed = random_int()
        elif random is False:
            seed = 0
        else:
            RaiseIt.value_error(random, ["True", "False"])

        rows_count = self.rows.count()
        if n < rows_count:
            # n/rows_count can return a number that represent less the total number we expect. multiply by 1.1 bo
            fraction = (n / rows_count) * 1.1
        else:
            fraction = 1.0

        return self.data.sample(False, fraction, seed=seed).limit(n)

    @staticmethod
    def stratified_sample(col_name, seed: int = 1):
        """
        Stratified Sampling
        :param col_name:
        :param seed:
        :return:
        """
        df = self
        fractions = df.select(col_name).distinct().withColumn("fraction", F.lit(0.8)).rdd.collectAsMap()
        df = df.stat.sampleBy(col_name, fractions, seed)
        return df

    @staticmethod
    def pivot(index, column, values):
        """
        Return reshaped DataFrame organized by given index / column values.
        :param self: Spark Dataframe
        :param index: Column to use to make new frame's index.
        :param column: Column to use to make new frame's columns.
        :param values: Column(s) to use for populating new frame's values.
        :return:
        """
        return self.groupby(index).pivot(column).agg(F.first(values))

    @staticmethod
    def melt(id_vars, value_vars, var_name="variable", value_name="value", data_type="str"):
        """
        Convert DataFrame from wide to long format.
        :param id_vars: column with unique values
        :param value_vars: Column names that are going to be converted to columns values
        :param var_name: Column name for vars
        :param value_name: Column name for values
        :param data_type: All columns must have the same type. It will transform all columns to this data type.
        :return:
        """

        df = self
        id_vars = val_to_list(id_vars)
        # Cast all columns to the same type
        df = df.cols.cast(id_vars + value_vars, data_type)

        vars_and_vals = [F.struct(F.lit(c).alias(var_name), F.col(c).alias(value_name)) for c in value_vars]

        # Add to the DataFrame and explode
        df = df.withColumn("vars_and_vals", F.explode(F.array(*vars_and_vals)))

        cols = id_vars + [F.col("vars_and_vals")[x].alias(x) for x in [var_name, value_name]]

        return df.select(*cols)

    @staticmethod
    def size():
        """
        Get the size of a dataframe in bytes
        :return:
        """

        def _to_java_object_rdd(rdd):
            """
            Return a JavaRDD of Object by unpickling
            It will convert each Python object into Java object by Pyrolite, whenever the
            RDD is serialized in batch or not.
            """
            rdd = rdd._reserialize(AutoBatchedSerializer(PickleSerializer()))
            return rdd.ctx._jvm.org.apache.spark.mllib.api.python.SerDe.pythonToJava(rdd._jrdd, True)

        if version.parse(Spark.instance.spark.version) < version.parse("2.4.0"):
            java_obj = _to_java_object_rdd(self.rdd)
            n_bytes = Spark.instance.sc._jvm.org.apache.spark.util.SizeEstimator.estimate(java_obj)
        else:
            # TODO: Find a way to calculate the dataframe size in spark 2.4
            n_bytes = -1

        return n_bytes


    def execute(self):
        self.data.cache()
        return self

    def compute(self):
        """
        This method is a very useful function to break lineage of transformations. By default Spark uses the lazy
        evaluation approach in processing data: transformation functions are not computed into an action is called.
        Sometimes when transformations are numerous, the computations are very extensive because the high number of
        operations that spark needs to run in order to get the results.

        Other important thing is that Apache Spark save task but not result of dataFrame, so tasks are
        accumulated and the same situation happens.

        :param self: Spark Dataframe
        :return:
        """

        self.data.cache().count()
        return self

    @staticmethod
    def query(sql_expression):
        """
        Implements the transformations which are defined by SQL statement. Currently we only support
        SQL syntax like "SELECT ... FROM __THIS__ ..." where "__THIS__" represents the
        underlying table of the input dataframe.
        :param self: Spark Dataframe
        :param sql_expression: SQL expression.
        :return: Dataframe with columns changed by SQL statement.
        """
        sql_transformer = SQLTransformer(statement=sql_expression)
        return sql_transformer.transform(self)

    @staticmethod
    def set_name(value=None):
        """
        Create a temp view for a data frame also used in the json output profiling
        :param value:
        :return:
        """
        self._name = value
        if not is_str(value):
            RaiseIt.type_error(value, ["string"])

        if len(value) == 0:
            RaiseIt.value_error(value, ["> 0"])

        self.createOrReplaceTempView(value)


    def partitions(self):
        """
        Return the dataframe partitions number
        :return: Number of partitions
        """
        df = self.data

        return df.to_spark().rdd.getNumPartitions()

    @staticmethod
    def partitioner():
        """
        Return the algorithm used to partition the dataframe
        :param self: Spark Dataframe
        :return:
        """
        return self.rdd.partitioner

    def repartition(self, n=None, *args, **kwargs):
        df = self.data
        df = df.repartition(n, *args, **kwargs)
        return self.new(df, meta=self.meta)

    def h_repartition(self, partitions_number=None, col_name=None):
        """
        Apply a repartition to a datataframe based in some heuristics. Also you can pass the number of partitions and
        a column if you need more control.
        See
        https://stackoverflow.com/questions/35800795/number-of-partitions-in-rdd-and-performance-in-spark/35804407#35804407
        https://medium.com/@adrianchang/apache-spark-partitioning-e9faab369d14
        https://umbertogriffo.gitbooks.io/apache-spark-best-practices-and-tuning/content/sparksqlshufflepartitions_draft.html
        :param self: Spark Dataframe
        :param partitions_number: Number of partitions
        :param col_name: Column to be used to apply the repartition id necessary
        :return:
        """

        if partitions_number is None:
            partitions_number = Spark.instance.parallelism * 4

        if col_name is None:
            df = self.repartition(partitions_number)
        else:
            df = self.repartition(partitions_number, col_name)
        return df

    @staticmethod
    def table_image(path, limit=10):
        """

        :param limit:
        :param path:
        :return:
        """

        css = absolute_path("/css/styles.css")

        imgkit.from_string(self.table_html(limit=limit, full=True), path, css=css)
        print_html("<img src='" + path + "'>")

    @staticmethod
    def isnotebook():
        """
        Detect you are in a notebook or in a terminal
        :return:
        """
        try:
            shell = get_ipython().__class__.__name__
            if shell == 'ZMQInteractiveShell':
                return True  # Jupyter notebook or qtconsole
            elif shell == 'TerminalInteractiveShell':
                return False  # Terminal running IPython
            else:
                return False  # Other type (?)
        except NameError:
            return False  # Probably standard Python interpreter

    def show(self):
        """
        :return:
        """
        self.data.show()

    @staticmethod
    def random_split(weights: list = None, seed: int = 1):
        """
        Create 2 random splited DataFrames
        :param weights:
        :param seed:
        :return:
        """
        if weights is None:
            weights = [0.8, 0.2]

        return self.randomSplit(weights, seed)

    def debug(self):
        """
        :return:
        """
        df = self.data
        print(df.rdd.toDebugString().decode("ascii"))

    def to_optimus_pandas(self):
        return PandasDataFrame(self.data.toPandas(), op=self.op)

    def to_pandas(self):
        return self.data.toPandas()




from typing import Any, Dict, Union
import unittest


class TestBase(unittest.TestCase):

    config: Dict[str, Union[str, int]]  = {"engine": "pandas", "n_partitions": 1}
    dict: Dict[Any, Any] = {}
    load: Dict[Any, Any]  = {}

    @classmethod
    def setUpClass(cls):

        if not cls.config:
            raise Exception("Please initialize device before running tests")

        import sys
        sys.path.append("..")
        from optimus import Optimus
        cls.op = Optimus(cls.config["engine"])
        print(f"optimus using {cls.op.engine}")

        if cls.dict:
            cls.df = cls.create_dataframe(cls.dict, force_data_types=True)
        elif cls.load:
            if isinstance(cls.load, str):
                cls.df = cls.load_dataframe(cls.load)
            else:
                cls.df = cls.load_dataframe(**cls.load)
        cls.post_create()

    @classmethod
    def create_dataframe(cls, data=None, **kwargs):
        if not "n_partitions" in kwargs and "n_partitions" in cls.config:
            kwargs["n_partitions"] = cls.config["n_partitions"]

        return cls.op.create.dataframe(data, **kwargs)

    @classmethod
    def load_dataframe(cls, path=None, type='file', **kwargs):
        if not "n_partitions" in kwargs and "n_partitions" in cls.config:
            kwargs["n_partitions"] = cls.config["n_partitions"]

        return getattr(cls.op.load, type)(path, **kwargs)

    @classmethod
    def post_create(cls):
        pass



from pyspark.ml import feature, classification
from pyspark.ml.classification import RandomForestClassifier, DecisionTreeClassifier, GBTClassifier
from pyspark.sql import functions as F
from pysparkling import *

# from optimus.infer import is_str, is_spark_dataframe
from optimus.helpers.check import is_spark_dataframe
from optimus.helpers.columns import parse_columns, name_col
from optimus.engines.base.ml.constants import STRING_TO_INDEX
from optimus.engines.spark.ml.encoding import string_to_index, vector_assembler
from optimus.engines.spark.spark import Spark
from optimus.infer import is_str

from optimus.engines.base.ml.models import BaseML

class ML(BaseML):
    @staticmethod
    def logistic_regression_text(df, input_col):
        """
        Runs a logistic regression for input (text) DataFrame.
        :param df: Pyspark dataframe to analyze
        :param input_col: Column to predict
        :return: DataFrame with logistic regression and prediction run.
        """

        if not is_spark_dataframe(df):
            raise TypeError("Spark dataframe expected")

        pl = feature.Tokenizer().setInputCol(input_col) | feature.CountVectorizer()
        ml = pl | classification.LogisticRegression()
        ml_model = ml.fit(df)
        df_model = ml_model.transform(df)
        return df_model, ml_model

    @staticmethod
    def random_forest(df, columns, input_col, **kwargs):
        """
        Runs a random forest classifier for input DataFrame.
        :param df: Pyspark dataframe to analyze.
        :param columns: List of columns to select for prediction.
        :param input_col: Column to predict.
        :return: DataFrame with random forest and prediction run.
        """

        columns = parse_columns(df, columns)

        data = df.select(columns)
        feats = data.columns
        feats.remove(input_col)

        df = string_to_index(df, input_cols=input_col)
        df = vector_assembler(df, input_cols=feats, output_col="features")

        model = RandomForestClassifier(**kwargs)
        df = df.cols.rename(name_col(input_col, STRING_TO_INDEX), "label")

        rf_model = model.fit(df)
        df_model = rf_model.transform(df)
        return df_model, rf_model

    @staticmethod
    def decision_tree(df, columns, input_col, **kwargs):
        """
        Runs a decision tree classifier for input DataFrame.
        :param df: Pyspark dataframe to analyze.
        :param columns: List of columns to select for prediction.
        :param input_col: Column to predict.
        :return: DataFrame with decision tree and prediction run.
        """

        if not is_spark_dataframe(df):
            raise TypeError("Spark dataframe expected")

        columns = parse_columns(df, columns)

        if not is_str(input_col):
            raise TypeError("Error, input column must be a string")

        data = df.select(columns)
        feats = data.columns
        feats.remove(input_col)

        df = string_to_index(df, input_cols=input_col)
        df = vector_assembler(df, input_cols=feats, output_col="features")

        model = DecisionTreeClassifier(**kwargs)

        df = df.cols.rename(name_col(input_col, STRING_TO_INDEX), "label")

        dt_model = model.fit(df)
        df_model = dt_model.transform(df)
        return df_model, dt_model

    @staticmethod
    def gbt(df, columns, input_col, **kwargs):
        """
        Runs a gradient boosting tree classifier for input DataFrame.
        :param df: Pyspark dataframe to analyze.
        :param columns: List of columns to select for prediction.
        :param input_col: Column to predict.
        :return: DataFrame with gradient boosting tree and prediction run.
        """

        if not is_spark_dataframe(df):
            raise TypeError("Spark dataframe expected")

        columns = parse_columns(df, columns)

        if not is_str(input_col):
            raise TypeError("Error, input column must be a string")

        data = df.select(columns)
        feats = data.columns
        feats.remove(input_col)

        df = string_to_index(df, input_cols=input_col)
        df = vector_assembler(df, input_cols=feats, output_col="features")

        model = GBTClassifier(**kwargs)

        df = df.cols.rename(name_col(input_col, STRING_TO_INDEX), "label")

        gbt_model = model.fit(df)
        df_model = gbt_model.transform(df)
        return df_model, gbt_model




# DataFrame = pd.DataFrame
from datetime import datetime

import numpy as np
import pandas as pd
import ibis

from optimus.engines.base.functions import BaseFunctions
from optimus.helpers.core import val_to_list


class IbisFunctions(BaseFunctions):

    _engine = ibis

    def to_float(self, series, *args):
        return series.cast("float64")

    def to_string(self, series):
        return series.cast("string")
    
    def to_string_accessor(self, series):
        return self.to_string(series)


    def title(self, series, *args):
        raise NotImplementedError("Not implemented yet")

    #
    # def replace(self, series, *args):
    #     return series.cast("string").upper()

    def count_zeros(self, *args):
        series = self.series
        return int((series.to_float().values == 0).sum())

    def kurtosis(self, series, *args):
        return self.to_float(series).kurt()

    def skew(self, series, *args):
        return self.to_float(series).skew()

    def exp(self, series, *args):
        return self.to_float(series).exp()

    def sqrt(self, series, *args):
        return self.to_float(series).sqrt()

    def radians(self):
        series = self.series
        return np.radians(series.to_float())

    def degrees(self):
        series = self.series
        return np.degrees(series.to_float())

    def ln(self, series, *args):
        return self.to_float(series).log()

    def log(self, series, base=10):
        return self.to_float(series).log() / np.log(base)

    def ceil(self, series, *args):
        return self.to_float(series).ceil()

    def sin(self):
        series = self.series
        return np.sin(series.to_float())

    def cos(self, series, *args):
        return self.to_float(series).cos()

    def tan(self):
        series = self.series
        return np.tan(series.to_float())

    def asin(self):
        series = self.series
        return np.arcsin(series.to_float())

    def acos(self):
        series = self.series
        return np.arccos(series.to_float())

    def atan(self):
        series = self.series
        return np.arctan(series.to_float())

    def sinh(self):
        series = self.series
        return np.arcsinh(series.to_float())

    def cosh(self):
        series = self.series
        return np.cosh(series.to_float())

    def tanh(self):
        series = self.series
        return np.tanh(series.to_float())

    def asinh(self):
        series = self.series
        return np.arcsinh(series.to_float())

    def acosh(self):
        series = self.series
        return np.arccosh(series.to_float())

    def atanh(self):
        series = self.series
        return np.arctanh(series.to_float())

    def clip(self, lower_bound, upper_bound):
        series = self.series
        return series.clip(lower_bound, upper_bound)

    def cut(self, bins):
        series = self.series
        return series.to_float(series).cut(bins, include_lowest=True, labels=list(range(bins)))

    def replace_chars(self, search, replace_by, ignore_case):
        series = self.series
        # if ignore_case is True:
        #     # Cudf do not accept re.compile as argument for replace
        #     # regex = re.compile(str_regex, re.IGNORECASE)
        #     regex = str_regex
        # else:
        #     regex = str_regex
        replace_by = val_to_list(replace_by)
        for i, j in zip(search, replace_by):
            series = series.astype(str).str.replace(i, j)
        return series

    def normalize_chars(self):
        series = self.series
        return series.str.normalize("NFKD").str.encode('ascii', errors='ignore').str.decode('utf8')

    def format_date(self, current_format=None, output_format=None):
        series = self.series
        return pd.to_datetime(series, format=current_format, errors="coerce").dt.strftime(output_format)

    def time_between(self, date_format=None):
        series = self.series
        return (pd.to_datetime(series, format=date_format,
                               errors="coerce").dt.date - datetime.now().date())



from optimus.helpers.functions import results_equal
from optimus.tests.base import TestBase


class TestDFPandas(TestBase):
    def test_dataframe_dict_auto(self):
        data = {'object': [1, "2", 3.0], "int": [1, 2, 3], "float": [1.0, 2.0, 3.5]}
        df = self.op.create.dataframe(data)
        result = df.cols.data_type()
        expected = {'object': 'object', 'int': 'int64', 'float': 'float64'}

        self.assertTrue(results_equal(result, expected))

    def test_dataframe_dict_forced(self):
        data = {('object', 'int64'): [1, "2", 3.0], ("int", 'float'): [1, 2, 3], ("float", 'object'): [1.0, 2.0, 3.5]}
        df = self.op.create.dataframe(data, force_data_types=True)
        result = df.cols.data_type()
        expected = {'object': 'int64', 'int': 'float64', 'float': 'object'}

        self.assertTrue(results_equal(result, expected))

    def test_dataframe_pandas(self):
        import pandas as pd
        data = pd.util.testing.makeMixedDataFrame()
        df = self.op.create.dataframe(data)
        result = df.cols.data_type()
        expected = {'A': 'float64', 'B': 'float64', 'C': 'object', 'D': 'datetime64[ns]'}

        self.assertTrue(results_equal(result, expected))

    def test_dataframe_dict_forced(self):
        data = {('object', 'int64'): [1, "2", 3.0], ("int", 'float'): [1, 2, 3], ("float", 'object'): [1.0, 2.0, 3.5]}
        df = self.op.create.dataframe(data, force_data_types=True)
        result = df.cols.data_type()
        expected = {'object': 'int64', 'int': 'float64', 'float': 'object'}

        self.assertTrue(results_equal(result, expected))


class TestDFDask(TestDFPandas):
    config = {'engine': 'dask', 'n_partitions': 1}


class TestDFPartitionDask(TestDFPandas):
    config = {'engine': 'dask', 'n_partitions': 2}




from abc import abstractmethod, ABC
from enum import Enum


class AbstractDriver(ABC):
    """Base class for all database implementations"""

    @abstractmethod
    def properties(self) -> Enum:
        """
        Returns an enum containing the following database properties:

            * name
            * port
            * java_class
            * table_name

        :return: en enum of type ``DriverProperties``
        """
        pass

    @abstractmethod
    def url(self, *args, **kwargs) -> str:
        """
        Returns a string connection to the underlying database implementation.
        :param kwargs: connection properties
        :return: the connection string
        """
        pass

    @abstractmethod
    def table_names_query(self, *args, **kwargs) -> str:
        """
        Returns the query for all table names in the database.
        :param kwargs: query parameters
        :return: a query for all the tables in a given database
        """
        pass

    @abstractmethod
    def table_name_query(self, *args, **kwargs) -> str:
        """
        Returns the query for a table name in the database.
        :param kwargs: query parameters
        :return: a query for all the tables in a given database
       """
        pass

    @abstractmethod
    def count_query(self, *args, **kwargs) -> str:
        """
        Returns the query for counting the rows in a given table.
        :param kwargs: query parameters
        :return: a query to count the number of rows in a table
        """
        pass

    @abstractmethod
    def primary_key_query(self, *args, **kwargs) -> str:
        """
        Returns the query the primary keys in a given table.
        :param kwargs: query parameters
        :return: a query to count the number of rows in a table
        """
        pass

    @abstractmethod
    def min_max_query(self, *args, **kwargs) -> str:
        """
        Returns the query with the min and max values in a columns from a given table.
        :param kwargs: query parameters
        :return: a query to count the number of rows in a table
        """
        pass



from optimus.helpers.functions import prepare_url_schema
from optimus.helpers.types import *
from optimus.engines.base.dask.io.jdbc import DaskBaseJDBC
from optimus.engines.base.io.properties import DriverProperties
from optimus.helpers.constants import Schemas
from optimus.helpers.raiseit import RaiseIt


class Connection:
    """
    Generic
    """

    def __init__(self, config=None, **kwargs):
        if not kwargs["base_url"].endswith("/"):
            kwargs["base_url"] = kwargs["base_url"] + "/"
        self._storage_options = kwargs
        self.options = config if config else kwargs

    def path(self, path):
        if self._storage_options["base_url"] and not path.startswith(
                (self._storage_options["base_url"], *Schemas.list())):
            path = self._storage_options["base_url"] + path

        return path

    @property
    def storage_options(self):
        storage_options = self._storage_options.copy()
        storage_options.pop("base_url")

        if not len(storage_options.keys()):
            storage_options = None
        else:
            storage_options = {k: storage_options[k] for k in storage_options.keys() if not k.startswith("_")}

        return storage_options

    def boto(self):
        return self.options


class S3(Connection):
    """
    Amazon S3
    """

    type = "s3"

    def __init__(self, endpoint_url=None, bucket=None, **kwargs):
        """
        endpoint_url: http(s)://<...>
        bucket
        anon: Whether access should be anonymous (default False)
        key: Key from configures Space Access Key
        secret: Secret from Space Access Key
        key, secret: For user authentication
        token: If authentication has been done with some other S3 client
        use_ssl: Whether connections are encrypted and secure (default True)
        client_kwargs: Dict passed to the boto3 client, with keys such as region_name or endpoint_url.
            Notice: do not pass the config option here, please pass it’s content to config_kwargs instead.
        config_kwargs: Dict passed to the s3fs.S3FileSystem, which passes it to the boto3 client’s config option.
        requester_pays: Set True if the authenticated user will assume transfer costs,
            which is required by some providers of bulk data
        default_block_size, default_fill_cache: These are not of particular interest to Dask users,
            as they concern the behaviour of the buffer between successive reads
        kwargs: Other parameters are passed to the boto3 Session object, such as profile_name,
            to pick one of the authentication sections from the configuration files referred to above (see here)


        """
        config = kwargs.copy()
        config["endpoint_url"] = endpoint_url
        config["bucket"] = bucket

        if endpoint_url is None:
            RaiseIt.value_error(endpoint_url, "")

        endpoint_url = prepare_url_schema(endpoint_url, Schemas.HTTPS.value, force=False)

        kwargs["client_kwargs"] = kwargs.get("client_kwargs", {})
        kwargs["client_kwargs"]["endpoint_url"] = kwargs["client_kwargs"].get("endpoint_url", endpoint_url)

        if not kwargs.get("base_url", False):
            if bucket:
                kwargs["base_url"] = prepare_url_schema(bucket, Schemas.S3.value, force=True)
            elif endpoint_url:
                kwargs["base_url"] = prepare_url_schema(endpoint_url, Schemas.S3.value, force=True)

        super().__init__(config, **kwargs)

    @property
    def boto(self):
        endpoint_url = prepare_url_schema(self.options['endpoint_url'], Schemas.HTTPS.value)
        # "region_name": S3_REGION
        return {"endpoint_url": endpoint_url,
                "aws_access_key_id": self.options.get("key"),
                "aws_secret_access_key": self.options.get("secret")}


class Local(Connection):
    """
    Local file system
    """

    def __init__(self, **kwargs):
        """
        supports base_url
        :param kwargs:
        """
        super().__init__(**kwargs)


class MAS(Connection):
    """
    Microsoft Azure Storage
    """

    def __init__(self, **kwargs):
        """
        Authentication for adl requires tenant_id, client_id and client_secret in the storage_options dictionary.
        Authentication for abfs requires account_name and account_key in storage_options.
        :param kwargs:
        """
        super().__init__(**kwargs)


class GCS(Connection):
    def __init__(self, host, port, user, kerb_ticket):
        config["host"] = host
        config["port"] = port
        config["user"] = user
        config["kerb_ticket"] = kerb_ticket
        super().__init__(config, host, port, user, kerb_ticket)


class HDFS(Connection):
    def __init__(self, **kwargs):

        config = kwargs.copy()

        if not kwargs.get("url", False):
            kwargs["url"] = "hdfs://" + kwargs["user"]

            if kwargs.get("password", False):
                kwargs["url"] += ":" + kwargs["password"]

            kwargs["url"] += "@" + kwargs["host"]

            if kwargs.get("port", False):
                kwargs["url"] += ":" + kwargs["port"]

        elif not kwargs["url"].startswith("hdfs://"):
            kwargs["url"] = "hdfs://" + kwargs["url"]

        kwargs["base_url"] = kwargs["url"]
        kwargs.pop("url")

        super().__init__(config, **kwargs)


class Connect:

    def __init__(self, op):
        self.op = op

    def mysql(self, host=None, database=None, user=None, password=None, port=None, schema="public") -> 'ConnectionType':
        return DaskBaseJDBC(host, database, user, password, port=port, driver=DriverProperties.MYSQL.value["name"],
                            schema=schema, op=self.op, sso=sso)

    def postgres(self, host=None, database=None, user=None, password=None, port=None, schema="public") -> 'ConnectionType':
        return DaskBaseJDBC(host, database, user, password, port=port, driver=DriverProperties.POSTGRESQL.value["name"],
                            schema=schema, op=self.op)

    def mssql(self, host=None, database=None, user=None, password=None, port=None, schema="public") -> 'ConnectionType':
        return DaskBaseJDBC(host, database, user, password, port=port, driver=DriverProperties.SQLSERVER.value["name"],
                            schema=schema, op=self.op)

    def redshift(self, host=None, database=None, user=None, password=None, port=None, schema="public") -> 'ConnectionType':
        return DaskBaseJDBC(host, database, user, password, port=port, driver=DriverProperties.REDSHIFT.value["name"],
                            schema=schema, op=self.op)

    def sqlite(self, host=None, database=None, user=None, password=None, port=None, schema="public") -> 'ConnectionType':
        return DaskBaseJDBC(host, database, user, password, port=port, driver=DriverProperties.SQLITE.value["name"],
                            schema=schema, op=self.op)

    def bigquery(self, host=None, database=None, user=None, password=None, port=None, schema="public", project=None,
                 dataset=None) -> 'ConnectionType':
        return DaskBaseJDBC(host, database, user, password, port=port, driver=DriverProperties.BIGQUERY.value["name"],
                            schema=schema, bigquery_project=project, bigquery_dataset=dataset, op=self.op)

    def presto(self, host=None, database=None, user=None, password=None, port=None, schema="public", catalog=None) -> 'ConnectionType':
        return DaskBaseJDBC(host, database, user, password, port=port, driver=DriverProperties.PRESTO.value["name"],
                            schema=schema, presto_catalog=catalog, op=self.op)

    def cassandra(self, host=None, database=None, user=None, password=None, port=None, schema="public", keyspace=None,
                  table=None) -> 'ConnectionType':
        return DaskBaseJDBC(host, database, user, password, port=port, driver=DriverProperties.CASSANDRA.value["name"],
                            schema=schema, cassandra_keyspace=keyspace, cassandra_table=table, op=self.op)

    def redis(self, host=None, database=None, user=None, password=None, port=None, schema="public") -> 'ConnectionType':
        return DaskBaseJDBC(host, database, user, password, port=port, driver=DriverProperties.REDIS.value["name"],
                            schema=schema, op=self.op)

    def oracle(self, host=None, database=None, user=None, password=None, port=None, schema="public",
               tns=None, service_name=None, sid=None) -> 'ConnectionType':
        return DaskBaseJDBC(host, database, user, password, port=port, driver=DriverProperties.ORACLE.value["name"],
                            schema=schema, oracle_tns=tns, oracle_service_name=service_name, oracle_sid=sid, op=self.op)

    @staticmethod
    def s3(**kwargs) -> 'ConnectionType':
        return S3(**kwargs)

    @staticmethod
    def local(**kwargs) -> 'ConnectionType':
        return Local(**kwargs)

    @staticmethod
    def hdfs(**kwargs) -> 'ConnectionType':
        return HDFS(**kwargs)

    @staticmethod
    def gcs(**kwargs) -> 'ConnectionType':
        """
        Google Cloud Storage
        :param kwargs:
        :return:
        """
        return GCS(**kwargs)

    @staticmethod
    def mas(**kwargs) -> 'ConnectionType':
        """
        Microsoft Azure Storage
        :param kwargs:
        :return:
        """
        return MAS(**kwargs)



import numpy as np
import pandas as pd
from sklearn import preprocessing

from optimus.engines.base.commons.functions import string_to_index, index_to_string, find
from optimus.engines.base.pandas.columns import PandasBaseColumns
from optimus.engines.base.dataframe.columns import DataFrameBaseColumns

DataFrame = pd.DataFrame


class Cols(PandasBaseColumns, DataFrameBaseColumns):
    def __init__(self, df):
        super().__init__(df)

    def _series_to_pandas(self, series):
        return series

    def find(self, cols="*", sub=None, ignore_case=False):
        """
        Find the start and end position for a char or substring
        :param cols:
        :param ignore_case:
        :param sub:
        :return:
        """
        df = self.root
        return find(df, cols, sub, ignore_case)

    def to_timestamp(self, cols="*", date_format=None, output_cols=None):
        raise NotImplementedError('Not implemented yet')

    def astype(self, cols="*", output_cols=None, *args, **kwargs):
        raise NotImplementedError('Not implemented yet')

    # NLP
    def stem_words(self, input_col):
        raise NotImplementedError('Not implemented yet')

    def string_to_index(self, cols="*", output_cols=None):
        df = self.root
        df.le = df.le or preprocessing.LabelEncoder()
        df = string_to_index(df, cols, output_cols, df.le)

        return df

    def index_to_string(self, cols="*", output_cols=None):
        df = self.root
        df.le = df.le or preprocessing.LabelEncoder()
        df = index_to_string(df, cols, output_cols, df.le)

        return df
